from unittest import case

from . import nacos_tools
from typing import (
    Any,
    Mapping,
    Optional,
    Union
)
import httpx
import logging
from mcp.server.models import InitializationOptions
import mcp.types as types
from mcp.server import NotificationOptions, Server
import mcp.server.stdio

logger = logging.getLogger('mcp_server_nacos')
logger.info("Starting Nacos MCP Server")

USER_AGENT = 'Nacos-MCP-Server:v0.1.0'

class Result:
    def __init__(self, code: int, message: str, data: Any):
        self.code = code
        self.message = message
        self.data = data

    def is_success(self) -> bool:
        return self.code == httpx.codes.OK

class NacosServer:
    def __init__(self, host: str, port: int, access_token:str):
        self.host = host
        self.port = port
        self.access_token = access_token

    def get(self, name:str, url: str, params: Any = None) -> str:
        url = f'http://{self.host}:{self.port}{url}'
        logger.debug(f'GET {url} with params {params}')
        result = self._request(url, params=params)
        if result is None:
            return "Unexpected error: None result handled."
        if result.is_success():
            return str(result.data)
        return f'Do {name} failed with message: {result.message}'

    def _request(self, url: str, params: Mapping[str, Optional[Union[str, int, float, bool]]] | None = None) -> Result | None:
        """Make a request to the Nacos API with proper error handling."""
        headers = {
            "User-Agent": USER_AGENT,
            "AccessToken": self.access_token
        }
        with httpx.Client() as client:
            try:
                response = client.get(url, headers=headers, timeout=30.0, params=params)
                if response.status_code == httpx.codes.OK:
                    return Result(response.status_code, response.text, response.json())
                if response.status_code in [httpx.codes.UNAUTHORIZED, httpx.codes.FORBIDDEN]:
                    return Result(response.status_code, "UnAuthorized request to Nacos, please make sure the accessToken is validated.", None)
                response.raise_for_status()
            except httpx.HTTPStatusError as e:
                return Result(e.response.status_code, str(e), None)
            except Exception as e:
                return Result(httpx.codes.INTERNAL_SERVER_ERROR, str(e), None)

async def main(host: str, port: int, access_token:str):
    logger.info(f"Starting Nacos MCP Server on {host}:{port} with token {access_token}")
    nacos = NacosServer(host, port, access_token)
    server = Server("mcp-nacos")

    @server.list_tools()
    async def handle_list_tools() -> list[types.Tool]:
        """List available tools"""
        return [
            nacos_tools.NacosListNamespacesTool(),
            nacos_tools.NacosListServices(),
            nacos_tools.NacosGetService(),
            nacos_tools.NacosListInstances(),
            nacos_tools.NacosListServiceSubscribers(),
            nacos_tools.NacosListConfigs(),
            nacos_tools.NacosGetConfig(),
            nacos_tools.NacosListConfigHistory(),
            nacos_tools.NacosGetConfigHistory(),
            nacos_tools.NacosListConfigListeners(),
            nacos_tools.NacosListListenedConfigs(),
        ]

    @server.call_tool()
    async def call_tool(name: str, arguments: dict) -> list[types.TextContent]:
        try:
            match name:
                case nacos_tools.NacosToolNames.LIST_NAMESPACES:
                    url = nacos_tools.NacosListNamespacesTool().url
                    result = nacos.get(name, url, arguments)
                    return [types.TextContent(type="text", text=result)]
                case nacos_tools.NacosToolNames.LIST_SERVICES:
                    url = nacos_tools.NacosListServices().url
                    result = nacos.get(name, url, arguments)
                    return [types.TextContent(type="text", text=result)]
                case nacos_tools.NacosToolNames.GET_SERVICE:
                    url = nacos_tools.NacosGetService().url
                    result = nacos.get(name, url, arguments)
                    return [types.TextContent(type="text", text=result)]
                case nacos_tools.NacosToolNames.LIST_INSTANCES:
                    url = nacos_tools.NacosListInstances().url
                    result = nacos.get(name, url, arguments)
                    return [types.TextContent(type="text", text=result)]
                case nacos_tools.NacosToolNames.LIST_SERVICE_SUBSCRIBERS:
                    url = nacos_tools.NacosListServiceSubscribers().url
                    result = nacos.get(name, url, arguments)
                    return [types.TextContent(type="text", text=result)]
                case nacos_tools.NacosToolNames.LIST_CONFIGS:
                    url = nacos_tools.NacosListConfigs().url
                    result = nacos.get(name, url, arguments)
                    return [types.TextContent(type="text", text=result)]
                case nacos_tools.NacosToolNames.GET_CONFIG:
                    url = nacos_tools.NacosGetConfig().url
                    result = nacos.get(name, url, arguments)
                    return [types.TextContent(type="text", text=result)]
                case nacos_tools.NacosToolNames.LIST_CONFIG_HISTORY:
                    url = nacos_tools.NacosListConfigHistory().url
                    result = nacos.get(name, url, arguments)
                    return [types.TextContent(type="text", text=result)]
                case nacos_tools.NacosToolNames.GET_CONFIG_HISTORY:
                    url = nacos_tools.NacosGetConfigHistory().url
                    result = nacos.get(name, url, arguments)
                    return [types.TextContent(type="text", text=result)]
                case nacos_tools.NacosToolNames.LIST_CONFIG_LISTENERS:
                    url = nacos_tools.NacosListConfigListeners().url
                    result = nacos.get(name, url, arguments)
                    return [types.TextContent(type="text", text=result)]
                case nacos_tools.NacosToolNames.LIST_LISTENED_CONFIGS:
                    url = nacos_tools.NacosListListenedConfigs().url
                    result = nacos.get(name, url, arguments)
                    return [types.TextContent(type="text", text=result)]
                case _:
                    raise ValueError(f"Unknown tool: {name}")

        except Exception as e:
            return [types.TextContent(type="text", text=str(e))]

    async with mcp.server.stdio.stdio_server() as (read_stream, write_stream):
        logger.info("Server running with stdio transport")
        await server.run(
            read_stream,
            write_stream,
            InitializationOptions(
                server_name="nacos",
                server_version="0.1.0",
                capabilities=server.get_capabilities(
                    notification_options=NotificationOptions(),
                    experimental_capabilities={},
                ),
            ),
        )



from . import server
import asyncio
import argparse


def main():
    """Main entry point for the package."""
    parser = argparse.ArgumentParser(description='Nacos Server')
    parser.add_argument('--host', default='localhost', help='Host to connect to Nacos Server')
    parser.add_argument('--port', default='8848', help='Port to connect to Nacos Server')
    parser.add_argument('--access_token', default='', help='AccessToken to connect to Nacos Server if auth enabled')

    args = parser.parse_args()
    asyncio.run(server.main(args.host, args.port, args.access_token))


# Optionally expose other important items at package level
__all__ = ["main", "server"]


from enum import Enum

import mcp.types as types


class NacosTool(types.Tool):
    url: str


class NacosListNamespacesTool(NacosTool):
    def __init__(self):
        super().__init__(
            name=NacosToolNames.LIST_NAMESPACES,
            description="Retrieves the list of namespaces in the current Nacos cluster.",
            inputSchema={
                "type": "object",
                "properties": {}
            },
            url="/nacos/v3/admin/core/namespace/list"
        )


class NacosListServices(NacosTool):
    def __init__(self):
        super().__init__(
            name=NacosToolNames.LIST_SERVICES,
            description="This interface retrieves the list of services under a specified namespace. The response format depends on the `withInstances` parameter:`withInstances=true`: Returns service details with instances (`ServiceDetailInfo` objects). `withInstances=false`: Returns service metadata without instances (`ServiceView` objects). **NOTE: ** When `withInstances=true`, The API may cost too much memory and networks, If Only want get instance list with little or one service, Suggest use `withInstances=false` with `List Service Instances`.",
            inputSchema={
                "type": "object",
                "properties": {
                    "pageNo": {"type": "int", "description": "The current page number, default is 1."},
                    "pageSize": {"type": "int", "description": "The size of services in each page, default is 100"},
                    "namespaceId": {"type": "string",
                                    "description": "The namespaceId of services, default is `public` if missing"},
                    "groupNameParam": {"type": "string",
                                       "description": "The groupName pattern of services, default null means all group if missing. if not null, server will search all service match groupName both prefix and subfix, such as: input `test`, groupName `test`, `atest`, `testb`, `atestb` will all matched"},
                    "serviceNameParam": {"type": "string",
                                         "description": "The serviceName pattern of services, default null means all service if missing.  if not null, server will search all service match serviceName both prefix and subfix, such as: input `test`, serviceName `test`, `atest`, `testb`, `atestb` will all matched"},
                    "ignoreEmptyService": {"type": "bool",
                                           "description": "Whether ignore the empty service in result, default is true"},
                    "withInstances": {"type": "bool",
                                      "description": "Whether contain instances under each services in result, recommend and default is false"},
                },
                "required": ["pageNo", "pageSize"],
            },
            url="/nacos/v3/admin/ns/service/list"
        )


class NacosGetService(NacosTool):
    def __init__(self):
        super().__init__(
            name=NacosToolNames.GET_SERVICE,
            description="This interface retrieves detailed information of a specified service, including metadata and clusters, not including instance list.",
            inputSchema={
                "type": "object",
                "properties": {
                    "namespaceId": {"type": "string",
                                    "description": "The namespaceId of services, default is `public` if missing"},
                    "groupName": {"type": "string",
                                  "description": "The groupName pattern of services, default is `DEFAULT_GROUP` if missing"},
                    "serviceName": {"type": "string",
                                    "description": "The serviceName pattern of services, required."}
                },
                "required": ["serviceName"],
            },
            url="/nacos/v3/admin/ns/service"
        )


class NacosListInstances(NacosTool):
    def __init__(self):
        super().__init__(
            name=NacosToolNames.LIST_INSTANCES,
            description="This interface retrieves the list of instances for a specified service.",
            inputSchema={
                "type": "object",
                "properties": {
                    "namespaceId": {"type": "string",
                                    "description": "The namespaceId of service, default is `public` if missing"},
                    "groupName": {"type": "string",
                                  "description": "The groupName pattern of service, default is `DEFAULT_GROUP` if missing"},
                    "serviceName": {"type": "string",
                                    "description": "The serviceName pattern of service, required."},
                    "clusterName": {"type": "string",
                                    "description": "The cluster name of instances in service, optional and default is null means match all cluster. If need match multiple cluster, use `,` to split like `cluster1,cluster2`"},
                },
                "required": ["serviceName"],
            },
            url="/nacos/v3/admin/ns/instance/list"
        )


class NacosListServiceSubscribers(NacosTool):
    def __init__(self):
        super().__init__(
            name=NacosToolNames.LIST_SERVICE_SUBSCRIBERS,
            description="This interface retrieves the list of subscribers for a specified service.",
            inputSchema={
                "type": "object",
                "properties": {
                    "pageNo": {"type": "int", "description": "The current page number, default is 1."},
                    "pageSize": {"type": "int", "description": "The size of subscribers in each page, default is 100"},
                    "namespaceId": {"type": "string",
                                    "description": "The namespaceId of service, default is `public` if missing"},
                    "groupName": {"type": "string",
                                  "description": "The groupName pattern of service, default is `DEFAULT_GROUP` if missing"},
                    "serviceName": {"type": "string",
                                    "description": "The serviceName pattern of service, required."},
                    "aggregation": {"type": "bool",
                                    "description": "Whether aggregation from whole cluster, if `false`, only get subscribers from requested node, default `true` if missing"}
                },
                "required": ["pageNo", "pageSize", "serviceName"],
            },
            url="/nacos/v3/admin/ns/service/subscribers"
        )


class NacosListConfigs(NacosTool):
    def __init__(self):
        super().__init__(
            name=NacosToolNames.LIST_CONFIGS,
            description="This interface retrieves the list of configurations under a specified namespace.",
            inputSchema={
                "type": "object",
                "properties": {
                    "pageNo": {"type": "int", "description": "The current page number, default is 1."},
                    "pageSize": {"type": "int", "description": "The size of configs in each page, default is 100"},
                    "namespaceId": {"type": "string",
                                    "description": "The namespaceId of configs, default is `public` if missing"},
                    "groupName": {"type": "string",
                                  "description": "The groupName pattern of configs, default null means all group. if not null, server will search all config match groupName. If `search=blur`, this parameter allow use `*` to do blur search, such as `*test` to do prefix and `test*` to do suffix"},
                    "dataId": {"type": "string",
                               "description": "The dataId pattern of configs, default null means all dataId. if not null, server will search all config match dataId. If `search=blur`, this parameter allow use `*` to do blur search, such as `*test` to do prefix and `test*` to do suffix"},
                    "type": {"type": "string",
                             "description": "The type of configs, default null means all type, if not null, server will search all config match type. Multiple type using `,` to split, such as `text,json`"},
                    "configTags": {"type": "string",
                                   "description": "The tags of configs, default null means all tags, if not null, server will search all config match tags. Multiple tags using `,` to split, such as `tag1,tag2`"},
                    "appName": {"type": "string",
                                "description": "The appName of configs, default null means all appName, if not null, server will search all config match appName."},
                    "search": {
                        "type": "string",
                        "description": "The search way of list configs, default is `blur` means blur search groupName and dataId. Or using `accurate` means accurate match with groupName and dataId."
                    }
                },
                "required": ["pageNo", "pageSize"],
            },
            url="/nacos/v3/admin/cs/config/list"
        )


class NacosGetConfig(NacosTool):
    def __init__(self):
        super().__init__(
            name=NacosToolNames.GET_CONFIG,
            description="This interface retrieves the details of the specified configuration.",
            inputSchema={
                "type": "object",
                "properties": {
                    "namespaceId": {"type": "string",
                                    "description": "The namespaceId of configs, default is `public` if missing"},
                    "groupName": {"type": "string",
                                  "description": "The groupName of config, Required."},
                    "dataId": {"type": "string",
                               "description": "The dataId of config, Required."}
                },
                "required": ["groupName", "dataId"],
            },
            url="/nacos/v3/admin/cs/config"
        )


class NacosListConfigHistory(NacosTool):
    def __init__(self):
        super().__init__(
            name=NacosToolNames.LIST_CONFIG_HISTORY,
            description="This interface retrieves the complete publish history of a configuration.",
            inputSchema={
                "type": "object",
                "properties": {
                    "pageNo": {"type": "int", "description": "The current page number, default is 1."},
                    "pageSize": {"type": "int",
                                 "description": "The size of config history records in each page, default is 100"},
                    "namespaceId": {"type": "string",
                                    "description": "The namespaceId of config, default is `public` if missing"},
                    "groupName": {"type": "string",
                                  "description": "The groupName pattern of config, required."},
                    "dataId": {"type": "string",
                               "description": "The dataId pattern of config, required."},
                },
                "required": ["pageNo", "pageSize", "groupName", "dataId"],
            },
            url="/nacos/v3/admin/cs/history/list"
        )


class NacosGetConfigHistory(NacosTool):
    def __init__(self):
        super().__init__(
            name=NacosToolNames.GET_CONFIG_HISTORY,
            description="This interface retrieves a specific historical change record of a configuration.",
            inputSchema={
                "type": "object",
                "properties": {
                    "namespaceId": {"type": "string",
                                    "description": "The namespaceId of configs, default is `public` if missing"},
                    "groupName": {"type": "string",
                                  "description": "The groupName of config, Required."},
                    "dataId": {"type": "string",
                               "description": "The dataId of config, Required."},
                    "nid": {"type": "long",
                            "description": "the actual id of config history record, Get from list config history api/tool, `id` field."},
                },
                "required": ["groupName", "dataId"],
            },
            url="/nacos/v3/admin/cs/history"
        )

class NacosListConfigListeners(NacosTool):
    def __init__(self):
        super().__init__(
            name=NacosToolNames.LIST_CONFIG_LISTENERS,
            description="This interface retrieves the list of listeners subscribed to a specific configuration.",
            inputSchema={
                "type": "object",
                "properties": {
                    "namespaceId": {"type": "string",
                                    "description": "The namespaceId of configs, default is `public` if missing"},
                    "groupName": {"type": "string",
                                  "description": "The groupName of config, Required."},
                    "dataId": {"type": "string",
                               "description": "The dataId of config, Required."},
                    "aggregation": {"type": "bool",
                                    "description": "Whether aggregation from whole cluster, if `false`, only get listeners from requested node, default `true` if missing"}
                },
                "required": ["groupName", "dataId"],
            },
            url="/nacos/v3/admin/cs/config/listener"
        )

class NacosListListenedConfigs(NacosTool):
    def __init__(self):
        super().__init__(
            name=NacosToolNames.LIST_LISTENED_CONFIGS,
            description="This interface lists the configurations subscribed to by a specific client IP address.",
            inputSchema={
                "type": "object",
                "properties": {
                    "namespaceId": {"type": "string",
                                    "description": "The namespaceId of configs, default is `public` if missing"},
                    "ip": {"type": "string",
                                  "description": "The client ip of config listeners, Required."},
                    "aggregation": {"type": "bool",
                                    "description": "Whether aggregation from whole cluster, if `false`, only get listeners from requested node, default `true` if missing"}
                },
                "required": ["ip"],
            },
            url="/nacos/v3/admin/cs/listener"
        )

class NacosToolNames(str, Enum):
    LIST_NAMESPACES = "list_namespaces",
    LIST_SERVICES = "list_services",
    GET_SERVICE = "get_service",
    LIST_INSTANCES = "list_service_instances",
    LIST_SERVICE_SUBSCRIBERS = "list_service_subscribers",
    LIST_CONFIGS = "list_configs",
    GET_CONFIG = "get_config",
    LIST_CONFIG_HISTORY = "list_config_history",
    GET_CONFIG_HISTORY = "get_config_history",
    LIST_CONFIG_LISTENERS = "list_config_listeners",
    LIST_LISTENED_CONFIGS = "list_listened_configs",
